/******************************************************************************

 This file is part of FDD3000 Emulator version 0.1.6

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>

 FDD300Emu is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>

******************************************************************************/

#ifndef DISKIMAGE_H
#define DISKIMAGE_H

#include <QLabel>
#include <QList>
#include <QImage>
#include <QPicture>
#include <QPixmap>
#include <QtGui>
#include <QEvent>
#include <QMenu>

enum dskFormat {
    FDD160K, /*SSDD*/
    FDD320K, /*DSDD (SS+flip)*/
    FDD640K  /*DSHD*/
};
enum imgFormat {
    IMG,
    DSK
};

enum sysName {
    sysNone,
    sysTOSv2,
    sysTOSv4,
    sysCPM,
    sysUnknown
};

enum saveMode {
    saveAuto=0,
    saveAsk,
    saveForce
};

enum dskError {
    dskOK,
    dskUnknownFormat,
    dskCannotOpen,
    dskLoadError,
    dskFileNotfound,
    dskDskNotSupported,
    dskWrongTrackSize,
    dskWrongGeometry,
    dskImportError,
    dskDoNotSave,
    dskSaveCanceled,
    dskCannotEject,
    dskTrackTooHigh,
    dskSectorTooHigh,
    dskSingleSide,
    dskReadOnly,
    dskNoDisk
};

class diskImage : public QLabel
{
    Q_OBJECT
public:
    explicit diskImage(QWidget *parent=0, QObject *gui = 0, int id=0);

private:
        const static quint8 trkHeader[];
        const static quint8 dskHeader[];
        const static quint8 diskTosA2[];

        QByteArray * image;
        QObject * logger;
        QString imageName;
        int trackJump;
        int trackOffset;
        int lastTrack;
        int format;
        int dirty;
        int loaded;
        int protect;
        int ID;
        int autosave;
        int side;
        int system;
        QString tosName;
//        int rotate;
private:
        int loadIMG(QString name);
        int loadDSK(QString name);
        int saveIMG(QString name);
        int saveDSK(QString name);
        void systemGuess();

public:
        ~diskImage();
        void setProtection(int ro);
        int openImage(void);
        int openImage(QString name);
        int closeImage(int mode);
        int closeImage(QString name);
        int ejectImage(int mode);
        int insertEmpty(int fmt);
        int swapSide(void);
        int loadSector(int side, int track, int sector, QByteArray * buf);
        int loadTrack(int side, int track, QByteArray * buf);
        int saveSector(int side, int track, int sector, QByteArray * buf);
        int saveTrack(int side, int track, QByteArray * buf);
//        QString tosName(void);
        int isLoaded(void);
        int isSwapable(void);
        int tracks(void);
        int isProtected(void);
        void setAutosave(int a);
        void rename(QString name);
signals:
        void message(QString s);

/********************* label stuff ********************/

private:
        QMenu   popup;
        QTimer  *spinTimer;
        QTimer  *rotateTimer;
        int     rotateCnt;
        int     rotate;
        static const QString icoNames[];
        QPixmap icons[8];

        enum iconSet {
            icoEmpty    = 0,
            icoLoaded,
            icoSideA,
            icoSideB,
            icoRotate
        };

private slots:
    void contextMenu(const QPoint&);
    void stopSpin(void);
    void rotateSpin(void);

signals:
        void spinStart();
        void spinStop();

public slots:
        void spin();


protected:
    void mousePressEvent(QMouseEvent *event);
    void paintEvent(QPaintEvent *event);
};

#endif // DISKIMAGE_H
