/******************************************************************************

 This file is part of FDD3000 Emulator version 0.1.6

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>

 FDD300Emu is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>

******************************************************************************/

#ifndef FDD3000_H
#define FDD3000_H

#include <QObject>
#include <QFile>
#include <QDebug>
#include <QThread>
#include <QTimer>
#include "sleeper.h"
#include "wd1772.h"
#include "diskImage.h"
#include "fddLink.h"
#include "z80/z80ex.h"
#include <stddef.h>

enum fd3Errors {
    fd3OK,
    fd3KeyfileOpen,
    fd3GetKeyfile,
    fd3CreateMessage,
    fd3DeleteMessage
};

class fdd3000 : public QObject
{
    Q_OBJECT
public:
    explicit fdd3000(QObject *parent = 0, diskImage ** drive = 0);
    ~fdd3000();

public:

private:
    class cpuThread : public QThread {
        protected:
            fdd3000 *fdd;
            Z80EX_CONTEXT *cpu;
            void run();
        public:
            cpuThread(fdd3000*f);
    };

private:
    cpuThread * z80Thread;
    Z80EX_CONTEXT *cpu;

    const static quint8 bootrom[];
    int romcs;
    quint8  ram[65536];
//    QTimer  sync;
    wd1772  *fdc;
    int     config;
    int     powered;
    int     faza;
    int     ipc;
    fddLink *link;

public:
    friend Z80EX_BYTE   memRead(Z80EX_CONTEXT *cpu, Z80EX_WORD addr, int m1_state, void *user_data);
    friend void         memWrite(Z80EX_CONTEXT *cpu, Z80EX_WORD addr, Z80EX_BYTE value, void *user_data);
    friend void         portWrite(Z80EX_CONTEXT *cpu, Z80EX_WORD port, Z80EX_BYTE value, void *user_data);
    friend Z80EX_BYTE   portRead(Z80EX_CONTEXT *cpu, Z80EX_WORD port, void *user_data);
    friend Z80EX_BYTE   intRead(Z80EX_CONTEXT *cpu, void *user_data);

private slots:
//    void syncSlot(void);
    void zxConnected();
    void zxDisconnected();

signals:
    void message(QString s);

public slots:
    int  powerOn(void);
    void powerOff(void);
    void fddReset(void);
};

Z80EX_BYTE  memRead(Z80EX_CONTEXT *cpu, Z80EX_WORD addr, int m1_state, void *user_data);
void        memWrite(Z80EX_CONTEXT *cpu, Z80EX_WORD addr, Z80EX_BYTE value, void *user_data);
void        portWrite(Z80EX_CONTEXT *cpu, Z80EX_WORD port, Z80EX_BYTE value, void *user_data);
Z80EX_BYTE  portRead(Z80EX_CONTEXT *cpu, Z80EX_WORD port, void *user_data);
Z80EX_BYTE  intRead(Z80EX_CONTEXT *cpu, void *user_data);


#endif // FDD3000_H
