/******************************************************************************

 This file is part of FDD3000 Emulator version 0.1.6

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>

 FDD300Emu is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>

******************************************************************************/


#include "fddLink.h"

#if defined (linux) || defined (__linux) || defined (__linux__)
const char * fddLink::shmId = "fdd.data";
const char * fddLink::fdd2tiE_Id = "fdd.fdd2ti.empty";
const char * fddLink::fdd2tiF_Id = "fdd.fdd2ti.full";
const char * fddLink::ti2fddE_Id = "fdd.ti2fdd.empty";
const char * fddLink::ti2fddF_Id = "fdd.ti2fdd.full";
const char * fddLink::fddAlive_Id = "fdd.alive";

fddLink::fddLink(QObject *parent) :
    QObject(parent)
{
    link=0;
    tiConnected=0;
    lastError=eOK;

    shm = shm_open(shmId, O_RDWR|O_CREAT, 0666);
    if(-1==shm){lastError=eShmOpenFailed; return; }
    if(-1==ftruncate(shm,sizeof(fdd2TiLink))) {
        lastError=eTruncate;
        close(shm);
        shm_unlink(shmId);
        return;
    }
    link=(fdd2TiLink*)mmap(0,sizeof(fdd2TiLink),PROT_READ|PROT_WRITE,MAP_SHARED,shm,0);
    if(link == MAP_FAILED) {lastError=eMmapFailed; close(shm); shm_unlink(shmId); return; }

    fddAlive=sem_open(fddAlive_Id,O_RDWR|O_CREAT,0666,1);
    if(fddAlive==SEM_FAILED) { lastError=eSemCreate; }
    fdd2tiE=sem_open(fdd2tiE_Id,O_RDWR|O_CREAT,0666,1);
    if(fddAlive==SEM_FAILED) { lastError=eSemCreate; }
    fdd2tiF=sem_open(fdd2tiF_Id,O_RDWR|O_CREAT,0666,0);
    if(fddAlive==SEM_FAILED) { lastError=eSemCreate; }
    ti2fddE=sem_open(ti2fddE_Id,O_RDWR|O_CREAT,0666,1);
    if(fddAlive==SEM_FAILED) { lastError=eSemCreate; }
    ti2fddF=sem_open(ti2fddF_Id,O_RDWR|O_CREAT,0666,0);
    if(fddAlive==SEM_FAILED) { lastError=eSemCreate; }
    
    if(lastError) { // prowizorka...
        sem_close(fddAlive);
        sem_close(ti2fddF);
        sem_close(ti2fddE);
        sem_close(fdd2tiF);
        sem_close(fdd2tiE);
        sem_unlink(fdd2tiE_Id);
        sem_unlink(fdd2tiF_Id);
        sem_unlink(ti2fddE_Id);
        sem_unlink(ti2fddF_Id);
        sem_unlink(fddAlive_Id);
        munmap(link,sizeof(fdd2TiLink));
        close(shm);
        shm_unlink(shmId);
    }

    semResetAll();

    sem_wait(fddAlive);
        link->fddAlive=1;
        link->tiAlive=0;
    sem_post(fddAlive);

//    thTimer  = new QThread(this);
    fddBeat  = new QTimer(this);
    fddBeat->setInterval(200);
    connect(fddBeat,SIGNAL(timeout()),this,SLOT(beat()));
    fddBeat->start();

//    fddBeat->moveToThread(thTimer);
//    thTimer->start();
}

fddLink::~fddLink(){
    sem_wait(fddAlive);
        tiConnected=link->tiAlive;
    sem_post(fddAlive);

    if(!lastError) {
        fddBeat->stop();
        delete fddBeat;
        semResetAll();
        sem_close(fddAlive);
        sem_close(ti2fddF);
        sem_close(ti2fddE);
        sem_close(fdd2tiF);
        sem_close(fdd2tiE);
        munmap(link,sizeof(fdd2TiLink));
        close(shm);
    }
    if(!tiConnected) {
        sem_unlink(fdd2tiE_Id);
        sem_unlink(fdd2tiF_Id);
        sem_unlink(ti2fddE_Id);
        sem_unlink(ti2fddF_Id);
        sem_unlink(fddAlive_Id);
        shm_unlink(shmId);
    }
//    thTimer->quit();
//    thTimer->wait(1);
//    delete thTimer;
}

void fddLink::semReset(sem_t * sem, qint32 val) {
    int i;
    int semval;
    for(i=0;i<50;i++) { sem_post(sem); }
    Sleeper::msleep(150);
    do {
        sem_wait(sem);
        sem_getvalue(sem,&semval);
    } while(semval>val);
}

void fddLink::semResetAll() {
    semReset(fddAlive,1);
    semReset(fdd2tiE,1);
    semReset(fdd2tiF,0);
    semReset(ti2fddE,1);
    semReset(ti2fddF,0);
}

void fddLink::beat() {
    int ti=tiConnected;
    sem_wait(fddAlive);
        ti=link->tiAlive;
    sem_post(fddAlive);

    if(ti&&(!tiConnected)) {
        emit(connected());
    }
    if((!ti)&&tiConnected) {
        emit(disconnected());
        semResetAll();
    }
    tiConnected=ti;
}

qint8 fddLink::read() {
    quint8 data=0;
    if (tiConnected) {
        sem_wait(ti2fddF);
        data=link->ti2fdd;
        sem_post(ti2fddE);
    }
    return data;
}

void fddLink::write(qint8 d) {
    if (tiConnected) {
        sem_wait(fdd2tiE);
        link->fdd2ti=d;
        sem_post(fdd2tiF);
    }
}
#endif

#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)

//const TCHAR* fddLink::shmId=TEXT("fdd.data");
/*
const TCHAR* fddLink::fdd2tiE_Id=TEXT("Global\\fdd.fdd2ti.empty");
const TCHAR* fddLink::fdd2tiF_Id=TEXT("Global\\fdd.fdd2ti.full");
const TCHAR* fddLink::ti2fddE_Id=TEXT("Global\\fdd.ti2fdd.empty");
const TCHAR* fddLink::ti2fddF_Id=TEXT("Global\\fdd.ti2fdd.full");
const TCHAR* fddLink::fddAlive_Id=TEXT("Global\\fdd.alive");
*/

const QString fddLink::fdd2tiE_Id="Global\\fdd.fdd2ti.empty";
const QString fddLink::fdd2tiF_Id="Global\\fdd.fdd2ti.full";
const QString fddLink::ti2fddE_Id="Global\\fdd.ti2fdd.empty";
const QString fddLink::ti2fddF_Id="Global\\fdd.ti2fdd.full";
const QString fddLink::fddAlive_Id="Global\\fdd.alive";


fddLink::fddLink(QObject *parent) :
    QObject(parent)
{

    QProcessEnvironment ev(QProcessEnvironment::systemEnvironment());

    shmfile=ev.value("TMP",ev.value("TEMP","."))+"\\fdd.data";
    //qDebug()<<"SHM file = "<<shmfile;

//    fdd2tiE_Id=TEXT("Global\\fdd.fdd2ti.empty");

    //    quint8 oldFdd=0;
    link=0;
    tiConnected=0;
    lastError=eOK;

    HANDLE shmf = CreateFile(shmfile.toStdWString().c_str(), GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL,NULL);
    if(shmf==NULL){
        lastError=eShmOpenFailed;
        qDebug() << "err shm file";
        return;
    }
    qDebug() << "file "<< shmfile << " open OK";
    shm = CreateFileMapping(
                shmf,//INVALID_HANDLE_VALUE,    // use paging file
                NULL,                    // default security
                PAGE_READWRITE,          // read/write access
                0,                       // maximum object size (high-order DWORD)
                256,                     // maximum object size (low-order DWORD)
                NULL);//TEXT("Global\\shmFdd"));                  // name of mapping object

    if(shm==NULL){
        lastError=eShmOpenFailed;
        qDebug() << "err shm";
        return;
    }
    qDebug() << "file mapped OK";
    link = (fdd2TiLink*)MapViewOfFile(shm,   // handle to map object
                                     FILE_MAP_ALL_ACCESS, // read/write permission
                                     0,
                                     0,
                                     sizeof(fdd2TiLink));

    if(link == NULL) {
        lastError=eMmapFailed;
        CloseHandle(shm);
        qDebug() << "err mmap";
        return;
    }
    qDebug() << "memory mapped OK";
    fddAlive=CreateSemaphore(NULL,1,1,fddAlive_Id.toStdWString().c_str());
    if(fddAlive==NULL) { lastError=eSemCreate;
        qDebug()<<"dup sem fddAlive Err:"<<GetLastError();
    }
    fdd2tiE=CreateSemaphore(NULL,1,1,fdd2tiE_Id.toStdWString().c_str());
    if(fdd2tiE==NULL) { lastError=eSemCreate;
        qDebug()<<"dup sem fdd2tiE Err:"<<GetLastError();
    }
    fdd2tiF=CreateSemaphore(NULL,0,1,fdd2tiF_Id.toStdWString().c_str());
    if(fdd2tiF==NULL) { lastError=eSemCreate;
        qDebug()<<"dup sem fdd2tiF Err:"<<GetLastError();
    }
    ti2fddE=CreateSemaphore(NULL,1,1,ti2fddE_Id.toStdWString().c_str());
    if(ti2fddE==NULL) { lastError=eSemCreate;
        qDebug()<<"dup sem ti2fddE Err:"<<GetLastError();
    }
    ti2fddF=CreateSemaphore(NULL,0,1,ti2fddF_Id.toStdWString().c_str());
    if(ti2fddF==NULL) { lastError=eSemCreate;
        qDebug()<<"dup sem ti2fddF Err:"<<GetLastError();
    }

    if(lastError) {
        UnmapViewOfFile(link);
        CloseHandle(shm);
        return;
    }
//    qDebug()<<"sem reset";
    semResetAll();
//    qDebug()<<"sem OK";
    WaitForSingleObject(fddAlive,INFINITE);
        link->fddAlive=1;
        link->tiAlive=0;
    ReleaseSemaphore(fddAlive,1,NULL);

//    thTimer  = new QThread(this);
    fddBeat  = new QTimer(this);
    fddBeat->setInterval(200);
    connect(fddBeat,SIGNAL(timeout()),this,SLOT(beat()));
    fddBeat->start();

//    fddBeat->moveToThread(thTimer);
//    thTimer->start();
}

fddLink::~fddLink(){  
    if(!lastError) {
        fddBeat->stop();
        delete fddBeat;
        CloseHandle(fddAlive);
        CloseHandle(ti2fddF);
        CloseHandle(ti2fddE);
        CloseHandle(fdd2tiF);
        CloseHandle(fdd2tiE);
        UnmapViewOfFile(link);
        CloseHandle(shm);
    }
    CloseHandle(shmf);
    QFile::remove(shmfile);
}

void fddLink::semReset(HANDLE sem, int val) {
//    qDebug() << "Sem: " << sem << "=" << ReleaseSemaphore(sem,3,NULL);
    usleep(1000);
//    int i;
//    for(i=0; i<10; i++) {
//        qDebug() << "Sem: " << sem << "=" << WaitForSingleObject(sem,0);
//    }
    while(!WaitForSingleObject(sem,0));
    if(val) { ReleaseSemaphore(sem,val,NULL); }
}


void fddLink::semResetAll() {
    semReset(fddAlive,1);
    semReset(fdd2tiE,1);
    semReset(fdd2tiF,0);
    semReset(ti2fddE,1);
    semReset(ti2fddF,0);

    /*
    ReleaseSemaphore(fddAlive,5,NULL);
    ReleaseSemaphore(fdd2tiE,5,NULL);
    ReleaseSemaphore(fdd2tiF,5,NULL);
    ReleaseSemaphore(ti2fddE,5,NULL);
    ReleaseSemaphore(ti2fddF,5,NULL);
    Sleeper::msleep(100);
    CloseHandle(fddAlive);
    CloseHandle(fdd2tiE);
    CloseHandle(fdd2tiF);
    CloseHandle(ti2fddE);
    CloseHandle(ti2fddF);
    fddAlive=CreateSemaphore(NULL,1,1,fddAlive_Id);
    fdd2tiE=CreateSemaphore(NULL,1,1,fdd2tiE_Id);
    fdd2tiF=CreateSemaphore(NULL,0,1,fdd2tiF_Id);
    ti2fddE=CreateSemaphore(NULL,1,1,ti2fddE_Id);
    ti2fddF=CreateSemaphore(NULL,0,1,ti2fddF_Id);
    */
}

void fddLink::beat() {
    int ti=tiConnected;
    WaitForSingleObject(fddAlive,INFINITE);
        ti=link->tiAlive;
    ReleaseSemaphore(fddAlive,1,NULL);

    if(ti&&(!tiConnected)) {
        emit(connected());
    }
    if((!ti)&&tiConnected) {
        emit(disconnected());
        semResetAll();
    }
    tiConnected=ti;
}

qint8 fddLink::read() {
    quint8 data=0;
    if (tiConnected) {
//        qDebug()<<"read";
        WaitForSingleObject(ti2fddF,INFINITE);
        data=link->ti2fdd;
        ReleaseSemaphore(ti2fddE,1,NULL);
    }
    return data;
}

void fddLink::write(qint8 d) {
    if (tiConnected) {
//        qDebug()<<"write";
        WaitForSingleObject(fdd2tiE,INFINITE);
        link->fdd2ti=d;
        ReleaseSemaphore(fdd2tiF,1,NULL);
    }
}
#endif
