/******************************************************************************

 This file is part of FDD3000 Emulator version 0.1.6

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>

 FDD300Emu is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>

******************************************************************************/

#ifndef FDDLINK_H
#define FDDLINK_H

#include <QObject>
//#include <QSharedMemory>
//#include <QSystemSemaphore>
#include <QProcess>
#include <QProcessEnvironment>
#include <QTimer>
#include <QDebug>
#include <QFile>
#include "sleeper.h"
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>

#if defined (linux) || defined (__linux) || defined (__linux__)
#include <sys/mman.h>
#include <sys/stat.h>        /* For mode constants */
#include <semaphore.h>
#endif

#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)
#include <windows.h>
#include <tchar.h>
#endif

typedef struct tFdd2TiLink {
    quint8  fdd2ti;
    quint8  ti2fdd;
    quint8  fddAlive;
    quint8  tiAlive;
} fdd2TiLink;

enum linkError {
    eOK,
    eShmOpenFailed,
    eMmapFailed,
    eTruncate,
    eFddExists,
    eSemCreate
};

class fddLink : public QObject
{
    Q_OBJECT
public:
    explicit    fddLink(QObject *parent = 0);
    ~fddLink();

private:
#if defined (linux) || defined (__linux) || defined (__linux__)
    static const char * shmId;
    static const char * fdd2tiE_Id;
    static const char * fdd2tiF_Id;
    static const char * ti2fddE_Id;
    static const char * ti2fddF_Id;
    static const char * fddAlive_Id;
    int        shm;
    sem_t      *fdd2tiF,*fdd2tiE,*ti2fddF,*ti2fddE,*fddAlive;
    void       semReset(sem_t * sem, qint32 val);
#endif
#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)
/*
    static const TCHAR* shmId;
    static const TCHAR* fdd2tiE_Id;
    static const TCHAR* fdd2tiF_Id;
    static const TCHAR* ti2fddE_Id;
    static const TCHAR* ti2fddF_Id;
    static const TCHAR* fddAlive_Id;
*/
static const QString shmId;
static const QString fdd2tiE_Id;
static const QString fdd2tiF_Id;
static const QString ti2fddE_Id;
static const QString ti2fddF_Id;
static const QString fddAlive_Id;

    HANDLE      shm;
    HANDLE      shmf;
    HANDLE      fdd2tiF,fdd2tiE,ti2fddF,ti2fddE,fddAlive;
    QString shmfile;
    void    semReset(HANDLE sem, int val);
#endif
    int        tiConnected;
    int        lastError;
    fdd2TiLink * link;
//    int      bitCnt;
//    QThread    *thTimer;
    QTimer     *fddBeat;

public:
    qint8       read();
    void        write(qint8 d);
//    int         tiReady();

signals:
    void message(QString s);
    void connected();
    void disconnected();

private slots:
    void beat();

public slots:
    void semResetAll();
};

#endif // FDDLINK_H
