/******************************************************************************

 This file is part of FDD3000 Emulator version 0.1.6

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>

 FDD300Emu is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>

******************************************************************************/

#include "gui.h"
#include "sleeper.h"
#include "ui_gui.h"

gui::gui(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::gui)
{
    int i;
    ui->setupUi(this);
    this->setMask(QBitmap(":/gfx/body_m.png"));

    fontFiles << ":ttf/berkeliumiidhr.ttf" << ":ttf/berkelium64.ttf"
              << ":ttf/petme642y.ttf" << ":ttf/shastonhi320.ttf"
              << ":ttf/shastonhi640.ttf" << ":ttf/zxdtpac.ttf";

    fontList <<"Berkelium 64"   <<"Berkelium II DHR"
             <<"Pet Me 64 2Y"   <<"Shaston Hi 320"
             <<"Shaston Hi 640" <<"ZX Academy";

    fontSizes << 10 << 12
              << 0  << 6
              << 10 << 0;

    for(i=0; i<fontFiles.count(); i++) {
        fontDB.addApplicationFont(fontFiles.at(i));
    }
    /** **/

    for(i=0; i<4; i++) {
        drive[i]=new diskImage(ui->rack,this,i);
        ui->hBoxLayout->addWidget(drive[i]);
        drive[i]->setAutosave(config.value("setup/autosave", 0).toInt());
    }
    drive[4]=new diskImage(NULL,this,4);

    /** **/

    setup = new setupdlg(this);
    setup->dlg->comboBox->addItems(fontList);
    setup->dlg->comboBox->setCurrentIndex(
                config.value("setup/font",0).toInt()
            );
    setup->dlg->diskName->setText(
                config.value("setup/disk", "").toString()
            );

    setup->dlg->checkBox->setChecked(config.value("setup/autosave", false).toBool());
    switch(config.value("setup/defaultaction", 0).toInt()) {
        case 0: setup->dlg->radioButton->setChecked(1);
                drive[0]->openImage(config.value("setup/disk", "").toString());
                break;
        case 1: setup->dlg->radioButton_2->setChecked(1);
                drive[0]->insertEmpty(FDD160K);
                break;
        default: setup->dlg->radioButton_3->setChecked(1);
                drive[0]->insertEmpty(FDD640K);
                break;
    }

    ui->infoLCD->setFont(
                QFont(
                    fontList.at(config.value("setup/font",0).toInt()),
                    fontSizes.at(config.value("setup/font",0).toInt())
                )
    );

    /** **/

    ui->infoLCD->appendPlainText("FDD3000 Emulator\n(c)Steev 2013,2014\n*************\nver 0.1.6a\n");
    QPoint pos = config.value("window/position", QPoint(200, 200)).toPoint();
    if( QApplication::desktop()->screenGeometry().contains(pos,true) ) { this->move(pos); }
    fdd3=new fdd3000(this,drive);
    fdd3->powerOn();/**/
    fdd3->fddReset();/**/
}

gui::~gui()
{
    int i;
    fdd3->powerOff();
    delete fdd3;
    fontDB.removeAllApplicationFonts();
    config.setValue("window/position", pos());
    config.sync();

    for(i=0; i<5; i++) {
        delete drive[i];
    }
    delete ui;
}

void gui::mousePressEvent(QMouseEvent *event) {
  if (event->button() == Qt::LeftButton) {
    lastPos = event->globalPos();
    if(QRect(10,105,30,20).contains(event->pos())) {
        fdd3->fddReset();
//        message("Reset...");
        return;
    }
    if(QRect(0,64,40,12).contains(event->pos())) {
//        message("bye");
//        fdd3->fddReset();
        fdd3->powerOff();
        Sleeper::msleep(300);
//        deleteLater();
        emit(QApplication::quit());
        return;
    }
    if(!QRect(185,15,475,120).contains(event->pos())) { down = true; };
  }

  if (event->button() == Qt::RightButton) {
      if(!QRect(185,15,475,120).contains(event->pos())) { reconfigure(); }
//      qDebug()<<setup->exec();
      return;
  }
//  QWidget::mousePressEvent(event);
}

void gui::mouseReleaseEvent(QMouseEvent *event) {
  down = false;
  QWidget::mouseReleaseEvent(event);
}

void gui::mouseMoveEvent(QMouseEvent *event) {
  if (down) {
    QPoint curPos = event->globalPos();
    if (curPos != lastPos) {
      QPoint diff = (lastPos - curPos);
      move(pos() - diff);
      lastPos = curPos;
    }
  }
  QWidget::mouseMoveEvent(event);
}

void gui::message(QString s) {
    ui->infoLCD->appendPlainText(s);
    update();
    qApp->processEvents();
}

void gui::reconfigure() {
    int i;
    if(!setup->exec()) return;
    config.setValue("setup/font", setup->dlg->comboBox->currentIndex());
    config.setValue("setup/disk", setup->dlg->diskName->text());
    config.setValue("setup/autosave", setup->dlg->checkBox->isChecked());
    if(setup->dlg->radioButton_2->isChecked()) { i=1; }
    else if (setup->dlg->radioButton_3->isChecked()) { i=2; }
    else { i=0; }
    config.setValue("setup/defaultaction",i);
    config.sync();

    ui->infoLCD->setFont(
        QFont(
            fontList.at(config.value("setup/font",0).toInt()),
            fontSizes.at(config.value("setup/font",0).toInt())
        )
    );
}
