/******************************************************************************

 This file is part of FDD3000 Emulator version 0.1.6

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>

 FDD300Emu is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>

******************************************************************************/

#include <QObject>
#include <QDebug>
#include "wd1772.h"
#include "sleeper.h"

#define VERBOSE_
#define VERBOSE2_

const quint16 wd1772::crc_fdc_table[] = {
  0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
  0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
  0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
  0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
  0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
  0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
  0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
  0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
  0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
  0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
  0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
  0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
  0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
  0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
  0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
  0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
  0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
  0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
  0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
  0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
  0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
  0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
  0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
  0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
  0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
  0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
  0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
  0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
  0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
  0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
  0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
  0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
};

wd1772::wd1772(QObject *parent, diskImage **img) :
    QObject(parent)
{
    fdd=img;
    buf=new QByteArray("");
    reset();
}

wd1772::~wd1772() {
    delete buf;
}

void wd1772::setConf(int val) {
    int i,romcs=(val&0x40);
    head=(val&0x80)?0:1;
    mfm=(val&0x20)?0:1; /*bez znaczenia...*/
    side=(val&0x10)?0:1;
    val&=0x0f;
    val^=0x0f;
    for(i=0; i<4;i++) { if ((0x01<<i)==val) { break; } }
    currentFdd=i;
#ifdef VERBOSE2
    qDebug()<<"Konfiguracja : ROM "<<(romcs?"N":"Y")
           <<", head "<<((head)?"down":"up")
           <<", format "<<((mfm)?"FMF":"MF")
           <<", side "<<((side)?"1":"0")
           <<", disk "<<currentFdd;
#endif
    (void)romcs;
}

void wd1772::cmdRestore(int c){
    (void)c;
    stepDir=stepOut;
    regTrack=currentTrack=0;
    regStatus=fdcIndex|fdcTrack0|fdcSpinUp;
    currentCommand=cmNone;
    intrq=1;
#ifdef VERBOSE
    qDebug()<<"wd1772 cmd restore";
#endif
}

void wd1772::cmdSeek(int c){
    int V=c&0x04;
    if(regData>regTrack) {stepDir=stepIn;}
    if(regData<regTrack) {stepDir=stepOut;}
    currentTrack=regTrack=regData;
    currentCommand=cmNone;
    intrq=1;
    if(V && (regTrack>=(unsigned int)fdd[currentFdd]->tracks())) { regStatus=fdcIndex|fdcSpinUp|fdcNotFound; return; }
    regStatus=fdcIndex|fdcSpinUp;
    if(!regTrack) { regStatus|=fdcTrack0; }
#ifdef VERBOSE
    qDebug()<<"wd1772 cmd seek (track "<<regData<<")";
#endif
}

void wd1772::cmdStep(int c){
    int V=c&0x04;
    int U=c&0x10;
    switch(c>>5) {
    case 0x02:  stepDir=stepIn; break;
    case 0x03:  stepDir=stepOut; break;
    }
    if(stepDir==stepOut){currentTrack--;} else {currentTrack++; }
    if(U) { regTrack=currentTrack; }
    intrq=1;
    if(V && (regTrack>=(unsigned int)fdd[currentFdd]->tracks())) { regStatus=fdcNotFound; return; }
    regStatus=fdcIndex|fdcSpinUp;
    if(!regTrack) { regStatus|=fdcTrack0; }
    currentCommand=cmNone;
#ifdef VERBOSE
    qDebug()<<"wd1772 cmd step";
#endif
}

void wd1772::cmdReadSector(int c){
    int M=c&0x10;
    if(M){currentCommand=cmReadSecMulti;} else {currentCommand=cmReadSec; }
    if(/*(!head)||*/fdd[currentFdd]->loadSector(side,currentTrack,regSector, buf)){
#ifdef VERBOSE
        qDebug()<<"wd1772 cmd read sector(d:" << currentFdd<<" tr:"<<currentTrack<<" sec:"<<regSector<<" NOT FOUND!)";
#endif
        currentCommand=cmNone;
        regStatus=fdcNotFound; intrq=1; return;
    }
    bufptr=0;
    regStatus=fdcDRQ|fdcBusy|fdcMotorOn;
    drq=1;
#ifdef VERBOSE
    qDebug()<<"wd1772 cmd read sector(d:" << currentFdd<<" tr:"<<currentTrack<<" sec:"<<regSector<<")";
#endif
}

void wd1772::cmdWriteSector(int c){
    int M=c&0x10;
    buf->clear();
    bufptr=0;
    if (!(head&fdd[currentFdd]->isLoaded())){       regStatus=fdcNotFound; intrq=1; return; }
    if (fdd[currentFdd]->isProtected()){ regStatus=fdcReadOnly; intrq=1; return; }
    if (M){currentCommand=cmWriteSecMulti;} else {currentCommand=cmWriteSec; }
    regStatus=fdcDRQ|fdcBusy|fdcMotorOn;
    drq=1;
#ifdef VERBOSE
    qDebug()<<"wd1772 cmd write sector(d:" << currentFdd<<" tr:"<<currentTrack<<" sec:"<<regSector<<")";
#endif
}

/* CRC-16-CCITT: G(x) = x^16 + x^12 + x^5 + 1  */
/* procedura podpatrzona u innych */
quint16 wd1772::getcrc( quint16 crc, quint8 data )
{
  return ( ( crc << 8 ) ^ crc_fdc_table[( ( crc >> 8 ) ^ data ) & 0xff] ) & 0xffff;
}

void wd1772::cmdReadAddress(int c){
    char addr[6];
    (void)c;
    if(!(head&fdd[currentFdd]->isLoaded())){ regStatus=fdcNotFound; intrq=1; return;}
    addr[0]=currentTrack&0xff;
    addr[1]=side&0xff;
    addr[2]=regSector&0xff;
    addr[3]=01;
    crc=0xb230;
    crc=getcrc(crc,addr[0]);
    crc=getcrc(crc,addr[1]);
    crc=getcrc(crc,addr[2]);
    crc=getcrc(crc,addr[3]);
    addr[4]=(crc&0xff);
    addr[5]=(crc>>8)&0xff;;
    buf->clear();
    buf->append(addr,6);
    currentCommand=cmReadAddr;
    bufptr=0;
    regStatus=fdcDRQ|fdcBusy|fdcMotorOn;
    drq=1;
#ifdef VERBOSE
    qDebug()<<"wd1772 cmd read address";
#endif
}

void wd1772::cmdInterrupt(int c){
    c&=0x0f;
    currentCommand=cmNone;
    buf->clear();
    bufptr=0;
    if(regStatus&fdcBusy) { regStatus&=(~fdcBusy); } else { regStatus=(fdd[currentFdd]->isLoaded())?fdcTrack0:0; }
    if(c&0x08) { intrq=1;}
    if(c&0x03) { qDebug()<< "cmdInterrupt z nie obsługiwaną opcją "<< c ; }
    /* TODO: int on index */  
#ifdef VERBOSE
    qDebug()<<"wd1772 cmd interrupt";
#endif
}

void wd1772::cmdReadTrack(int c){
    (void)c;
    /* TODO */
    qDebug()<< "cmdReadTrack : TODO";
    intrq=2;
}

void wd1772::cmdWriteTrack(int c){
    (void)c;
    qDebug()<< "wd1772 cmd write track";
    currentCommand=cmWriteTrack;
    buf->clear();
    bufptr=0;
    regStatus=fdcBusy|fdcDRQ|fdcMotorOn;
    drq=1;
}

void wd1772::command(int c) {
    int type = c&0xf0;
    currentCommand=cmNone;
    regStatus&=(~(fdcMotorOn|fdcIndex|fdcCRCError|fdcNotFound|fdcReadOnly));
    switch(type){
    case 0x00:  cmdRestore(c); break;
    case 0x10:  cmdSeek(c); break;
    case 0x20:
    case 0x30:
    case 0x40:
    case 0x50:
    case 0x60:
    case 0x70: cmdStep(c);break;
    case 0x80:
    case 0x81: cmdReadSector(c); break;
    case 0xA0:
    case 0xB0: cmdWriteSector(c); break;
    case 0xC0: cmdReadAddress(c); break;
    case 0xD0: cmdInterrupt(c); break;
    case 0xE0: cmdReadTrack(c); break;
    case 0xF0: cmdWriteTrack(c); break;
    }
}

int wd1772::readData() {
    if( (currentCommand!=cmReadSecMulti)&&
        (currentCommand!=cmReadSec)&&
        (currentCommand!=cmReadAddr)) { return regData; }
    int val = *(buf->data()+bufptr)&0xff;
    bufptr++;
    if(bufptr>=(unsigned int)buf->size()) {
        switch(currentCommand){
        case cmReadSec:
        case cmReadAddr: intrq=1; regStatus&=(~(fdcBusy|fdcDRQ));drq=0; currentCommand=cmNone; break;
        case cmReadSecMulti: regSector++; cmdReadSector(0x10); break;
        }
    }
    return val;
}

int wd1772::formatTrack(QByteArray * buf) {
    qint32 marker=0;
    quint8  trk=0,sid=0,sec=0;
    QByteArray track;
    QByteArray hdrMarker("\0\0\xf5\xf5\xf5\xfe",6);
    QByteArray dtaMarker("\0\0\xf5\xf5\xf5\xfb",6);

    while(-1<(marker=buf->indexOf(hdrMarker,marker))){
        trk=buf->at(marker+6);
        sid=buf->at(marker+7);
        sec=buf->at(marker+8);
        qDebug()<<"hdmarker t:"<<trk<<" s:"<<sid<<" sec:"<<sec;
        marker=buf->indexOf(dtaMarker,marker);
        track.append(buf->mid(marker+6,256));
        marker+=256;
    }
    if(sec==15) { return fdd[currentFdd]->saveTrack(sid,trk,&track); }
    else return fdcLostData;
}

void wd1772::writeData(int val) {
    int err;
    if( (currentCommand!=cmWriteSecMulti)&&(currentCommand!=cmWriteSec)&&(currentCommand!=cmWriteTrack)) { return; }
    buf->append((char)(val&0xff));
    if(currentCommand==cmWriteTrack){
//        qDebug()<<QString("[0x%1]=0x%2").arg(buf->length(),4,16,QChar('0')).arg(val&0xff,2,16,QChar('0'));
        if(buf->size()>=0x1570) {
            drq=0; intrq=1;
            regStatus&=(~(fdcBusy|fdcDRQ));
            if(formatTrack(buf)) {
                regStatus|=fdcLostData;
            }
        }
        return;
    }
    if(buf->size()>=256) {
        err=fdd[currentFdd]->saveSector(side,currentTrack,regSector,buf);
        if (err) {
            intrq=1; regStatus=fdcNotFound; currentCommand=cmNone; return; }
        buf->clear();
        if( currentCommand!=cmWriteSecMulti) {
            intrq=1;
            regStatus&=(~(fdcBusy|fdcDRQ));
            drq=0;
            currentCommand=cmNone;
            return;
        }
    }
}

void wd1772::writePort(int port, int val) {
    switch (port) {
    case fdpCommand:    regCommand=val; intrq=0; command(val); return;
    case fdpTrack:      regTrack=currentTrack=val; return;
    case fdpSector:     regSector=val; return;
    case fdpData:       regData=val; writeData(val); return;
    }
}

int  wd1772::readPort(int port) {
    int val;
    switch (port) {
    case fdpStatus:     val=regStatus;
                        if(!fdd[currentFdd]->isLoaded()) { val|=fdcNotReady;}
                        regStatus&=(fdcBusy|fdcNotReady);
                        return val;
    case fdpTrack:      return regTrack;
    case fdpSector:     return regSector;
    case fdpData:       regData=readData(); return regData;
    }
    return -1;
}

void wd1772::reset(void){
    regCommand=regData=regSector=regStatus=regTrack=0;
    head=mfm=side=stepDir=drq=intrq=currentFdd=currentTrack=0;
    buf->clear();
    bufptr=0;
}
