/******************************************************************************

 This file is part of FDD3000 Emulator version 0.1.6

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>

 FDD300Emu is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>

******************************************************************************/

#ifndef WD1772_H
#define WD1772_H

#include <QObject>
#include "diskImage.h"

enum fdPorts {
    fdpCommand=0,
    fdpStatus=0,
    fdpTrack,
    fdpSector,
    fdpData
};

enum fdCommand {
    cmNone,
    cmRestore,
    cmSeek,
    cmStep,
    cmStepIn,
    cmStepOut,
    cmReadSec,
    cmReadSecMulti,
    cmWriteSec,
    cmWriteSecMulti,
    cmReadAddr,
    cmReadTrack,
    cmWriteTrack,
    cmForceint
};

enum wdFlags {
    fdcBusy     = 0b00000001,
    fdcIndex    = 0b00000010,
    fdcDRQ      = 0b00000010,
    fdcTrack0   = 0b00000100,
    fdcLostData = 0b00000100,
    fdcCRCError = 0b00001000,
    fdcNotFound = 0b00010000,
    fdcRecType  = 0b00100000,
    fdcSpinUp   = 0b00100000,
    fdcReadOnly = 0b01000000,
    fdcMotorOn  = 0b10000000,
    fdcNotReady = 0b10000000
};

enum wdStep {
    stepIn,
    stepOut,
    stepDir
};

class wd1772 : public QObject
{
    Q_OBJECT
public:
    explicit wd1772(QObject *parent = 0, diskImage ** img=0);

private:
    const static quint16 crc_fdc_table[];
    QByteArray * buf;
    unsigned int bufptr;
    unsigned int regCommand;
    unsigned int regTrack;
    unsigned int regSector;
    unsigned int regData;
    unsigned int regStatus;
    unsigned int head;
    unsigned int mfm;
    unsigned int side;
    unsigned int currentFdd;
    unsigned int currentTrack;
    unsigned int currentCommand;
    unsigned int stepDir;
    unsigned short crc;
public:
    unsigned int drq;
    unsigned int intrq;

private:
    void command(int c);
    void cmdRestore(int c);
    void cmdSeek(int c);
    void cmdStep(int c);
    void cmdReadSector(int c);
    void cmdWriteSector(int c);
    void cmdReadAddress(int c);
    void cmdInterrupt(int c);
    void cmdReadTrack(int c);
    void cmdWriteTrack(int c);
    int  readData(void);
    void writeData(int val);
    int formatTrack(QByteArray * buf);
    quint16 getcrc( quint16 crc, quint8 data);
    diskImage ** fdd;
public:
    ~wd1772();
    void setConf(int val);
    void writePort(int port, int val);
    int  readPort(int port);
    void reset(void);
    int  isInt(void) { return intrq;}
    int  isDRQ(void) { return drq;  }
signals:
    
public slots:
    
};

#endif // WD1772_H
